from __future__ import annotations

import contextlib
import functools
import json
import logging
from typing import Any, Callable, Dict, List, Optional

from supervisor.manifest import StepStatus
from supervisor.recorder import Recorder
from supervisor.reporting.record import Record
from supervisor.reporting.report_builder import ReportBuilder, StepReportBuilder
from supervisor.reporting.status import Status
from supervisor.reporting.timer import Timer


def create_step_decorator(
    report_builder: ReportBuilder, recorder: Recorder
) -> Callable:
    """
    A decorator to mark a function as the implementation of a step in a
    digital worker's manifest.

    Note that the return type here is `Any`, because proper type hinting
    for decorators is not yet obvious. PEP 612 specifies a way to type
    hint decorators starting in 3.10, but this project supports >= 3.7. See
    more here: https://stackoverflow.com/a/68290080/2597913.

    Returns:
        Any: A decorator to attach to functions.
    """

    # The decorator to be returned as the property
    # This handles the inputs to the decorator itself
    def returned_decorator(*id_args):
        """
        To include a `supervisor_record_id`, pass it to the decorated function
        as a kwarg. For example,
        ```
            @step(1)
            def my_func(*args, supervisor_record_id: str):
                do stuff
                ...

            my_func(supervisor_record_id="my_record_id")
        ```
        """
        step_id = _step_id_from_args(id_args)

        # The decorator to grab the function callable itself
        def inner_decorator(fn):
            # And the wrapper around the function to call it with its
            # args and kwargs arguments
            @functools.wraps(fn)
            def wrapper(*fn_args, **fn_kwargs):
                return _run_wrapped_func(
                    fn, step_id, report_builder, recorder, *fn_args, **fn_kwargs
                )

            return wrapper

        return inner_decorator

    return returned_decorator


def _run_wrapped_func(
    fn: Callable,
    step_id: str,
    report_builder: ReportBuilder,
    recorder: Recorder,
    *fn_args: List,
    **fn_kwargs: Dict,
) -> Any:
    """
    Runs `fn` with the given args `args` and `kwargs`, times how long it
    takes to run, and records the execution of this function under `step_id`
    in the work report.

    Args:
    fn (Callable): The function to run.
    step_id (str): The ID of the step representing the function to
    execute.
    *args (List): The input arguments to the function.
    **kwargs (Dict): The input keyword arguments to the function.

    Returns:
    Any: Whatever is returned by executing the function
    """

    """Set up the step"""
    caught_exception: Optional[Exception] = None
    step_final_status: StepStatus

    # Time the function
    func_timer = Timer()
    func_timer.start()

    """Run the step"""
    try:
        fn_result = fn(*fn_args, **fn_kwargs)
        step_final_status = StepStatus.COMPLETE
        record_final_status = Status.SUCCEEDED
    except Exception as ex:
        fn_result = None
        caught_exception = ex
        step_final_status = StepStatus.ERROR
        record_final_status = Status.FAILED
    timer_result = func_timer.end()

    """Step teardown"""
    # Combine the linear args and keyword args into one dict
    args_as_dict = {"": fn_args} if fn_args else {}
    all_inputs = {**args_as_dict, **fn_kwargs}

    safe_args = to_safe_jsonable(all_inputs)
    safe_outputs = to_safe_jsonable(fn_result)

    # this checks if a record id was passed to the function and sets record_id
    # as none if it is not included in the kwargs
    record_id: str = fn_kwargs.get("supervisor_record_id")
    record = Record(record_id, record_final_status) if record_id else None

    # Create a new step report
    new_report = StepReportBuilder(
        step_id=step_id,
        start_time=timer_result.start,
        end_time=timer_result.end,
        duration=timer_result.duration,
        status=step_final_status,
        args=safe_args,
        outputs=safe_outputs,
        message_log=recorder.messages,
        data_log=recorder.data,
        record=record,
    )
    report_builder.workflow.append(new_report)

    # Reset the recorder for the next step
    recorder.messages = []
    recorder.data = []

    """Passthrough the step's return value or raised exception"""
    # If the step failed and raised an exception, raise that instead
    # of returning None
    if caught_exception:
        raise caught_exception

    # Passthrough the function's returned data back up
    return fn_result


def _step_id_from_args(args: tuple) -> str:
    """
    Joins decorator number args into a step ID string.

    Args:
    args (tuple): The arguments to the decorator.

    Returns:
    str: The step ID.
    """
    return ".".join([str(arg) for arg in args])


def to_safe_jsonable(value: Any) -> Any:
    """
    Returns a version of `value` that can be converted into JSON
    format using the `json` library.

    Args:
        value: The value to ensure is safe to JSON serialize.

    Returns:
        Any: The JSON safe value
    """
    # If the standard JSON library knows how to encode the object,
    # go with that
    with contextlib.suppress(TypeError):
        json.dumps(value)
        return value
    # Otherwise, check if the object has a .__json__() method
    json_attribute = getattr(value, "__json__", None)
    if callable(json_attribute):
        try:
            return value.__json__()
        except Exception:
            # [CX-234] If the call fails, warn and continue trying other options
            logging.warning("__json__ method raised exception", exc_info=True)

    # Use the raw string as a last resort
    return str(value)



"""
Default instances for dynamic runs.

Typically, developers will only run one digital worker on a Python
runtime, and it is tricky to initialize an object and pass that object
instance down into other files in a project for a custom decorator.
This package provides a set of default instances of objects and contexts
for dynamic runs.

This file initializes a shared `ReportBuilder` and `Recorder` for
individual steps and sub steps to share, then creates a default
`MainContext` named `supervise`, a decorator to annotate steps
called `step`, and a `StepContext` named `substep`.
"""

import pathlib
import warnings
from typing import Optional, Union

from supervisor.main_context import MainContext
from supervisor.recorder import Recorder
from supervisor.reporting.report_builder import ReportBuilder
from supervisor.step_context import StepContext
from supervisor.step_decorator_factory import create_step_decorator

report_builder = ReportBuilder()
#: A shared `ReportBuilder` for steps and sub-steps to add step reports to
recorder = Recorder()
#: A shared `Recorder` for steps and sub-steps to record messages to

step = create_step_decorator(report_builder, recorder)
#: Use this decorator on your own functions to mark what workflow step each
#: Python function is matched to.

fail_step = report_builder.fail_step
set_step_status = report_builder.set_step_status


#: Expose the report builder's ability to override a step's status as a
#: top-level call.


# noinspection PyPep8Naming
class step_scope(StepContext):
    def __init__(self, *id_args, record_id: Optional[str] = None):
        """
        A default `StepContext` that uses the root level `report_builder` and
        `recorder`.

        Args:
            *id_args: The list of integers that represent the step ID.
            record_id: An optional ID of the record being actively processed
        """
        super().__init__(report_builder, recorder, *id_args, record_id=record_id)


# noinspection PyPep8Naming
class substep(step_scope):
    """The deprecated version of `step_scope`."""

    def __init__(self, *id_args):
        warnings.simplefilter("once", DeprecationWarning)
        warnings.warn(
            f"`{self.__class__.__name__}` has been renamed to `step_scope`. "
            f"Please use that class name instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)
        super().__init__(*id_args)


# noinspection PyPep8Naming
class supervise(MainContext):
    def __init__(
        self,
        manifest: Union[str, pathlib.Path] = "manifest.yaml",
        output_dir: Union[str, pathlib.Path] = "output/",
        *args,
        **kwargs,
    ):
        """
        A default `MainContext` that uses the root level `report_builder` and
        `recorder`.

        Args:
            *args: Extra arguments to the `MainContext` constructor.
            **kwargs: Extra keyword arguments to the `MainContext` constructor.
            manifest (str): The digital worker's manifest definition
            output_dir (str): Where the work report and drift report will
                be written to
        """
        super().__init__(
            report_builder=report_builder,
            recorder=recorder,
            manifest=manifest,
            output_dir=output_dir,
            *args,
            **kwargs,
        )



from dataclasses import dataclass

from supervisor.reporting.status import Status


@dataclass
class Record:
    record_id: str
    status: Status

    def __json__(self):
        return {
            "id": self.record_id,
            "status": self.status.value,
        }



from __future__ import annotations

import warnings
from types import TracebackType
from typing import Optional, Type, Union

from supervisor.manifest import StepStatus
from supervisor.recorder import Recorder
from supervisor.reporting.record import Record
from supervisor.reporting.report_builder import ReportBuilder, StepReportBuilder
from supervisor.reporting.status import Status
from supervisor.reporting.timer import Timer


class StepContext:
    def __init__(
        self,
        builder: ReportBuilder,
        record: Recorder,
        *id_args,
        record_id: Optional[str] = None,
    ):
        """
        A context manager for a step that is running inside another step.
        This is an alternative to `@step` decorator when you don't want
        to write an entire function for a step.

        Args:
            builder: Where the step report will be written.
            record: Where messages and data logs will be written.
            *id_args: The numbers of the step, such as "1, 1"
            record_id: An optional ID of the record being actively processed
        """
        self.uuid = ".".join([str(n) for n in id_args])
        self.report_builder = builder
        self.recorder = record
        self.timer = Timer()
        self._status_override: Optional[StepStatus] = None
        self.record_id: Optional[str] = record_id
        self._record_status_override: Optional[StepStatus] = None

    def __enter__(self):
        """
        Logic for when this context is first started.

        Returns:
            MainContext: This instance.
        """
        self.timer.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """
        Runs when the context is about to close, whether caused
        by a raised Exception or now.

        Returns:
            bool: True if the parent caller should ignore the
                Exception raised before entering this function
                (if any), False otherwise.
        """
        timed_info = self.timer.end()
        if self._status_override:
            step_status = self._status_override
        else:
            step_status = StepStatus.ERROR if exc_type else StepStatus.COMPLETE

        if self._record_status_override:
            record_status = self._record_status_override
        else:
            record_status = Status.FAILED if exc_type else Status.SUCCEEDED

        record = Record(self.record_id, record_status) if self.record_id else None

        new_report = StepReportBuilder(
            step_id=self.uuid,
            start_time=timed_info.start,
            end_time=timed_info.end,
            duration=timed_info.duration,
            status=step_status,
            args={},
            outputs=None,
            message_log=self.recorder.messages,
            data_log=self.recorder.data,
            record=record,
        )

        self.report_builder.workflow.append(new_report)
        return False

    def error(self) -> None:
        """
        Sets the status of this step to `StepStatus.ERROR` in its `StepReport`.
        """
        self.set_status(StepStatus.ERROR)

    def set_status(self, status: Union[str, StepStatus]) -> None:
        """Override a step to be in the status of `status`"""
        # Convert the status to the correct type if necessary
        safe_status = StepStatus(status)
        self._status_override = safe_status

    def set_record_status(self, status: Union[str, Status]) -> None:
        """Override a step record to be in the status of `status`"""
        # Convert the status to the correct type if necessary
        if not self.record_id:
            warnings.warn(
                "Setting a record status for a step without "
                "a record ID will have no effect",
                UserWarning,
            )
        safe_status = Status(status)
        self._record_status_override = safe_status


if __name__ == "__main__":
    report_builder = ReportBuilder()
    recorder = Recorder()

    substep = StepContext

    with substep(report_builder, recorder, 1) as s:
        print("hello world")

        with substep(report_builder, recorder, 1, 1) as s2:
            print("inner step")

    print(report_builder.workflow)



from __future__ import annotations

import datetime
import logging
import pathlib
from types import TracebackType
from typing import Callable, Optional, Type, Union

import yaml

from supervisor.manifest import Manifest
from supervisor.recorder import Recorder
from supervisor.reporting.report import WorkerStatus
from supervisor.reporting.report_builder import ReportBuilder

logger = logging.getLogger(__name__)


class MainContext:
    def __init__(
        self,
        report_builder: ReportBuilder,
        recorder: Recorder,
        manifest: Union[str, pathlib.Path],
        output_dir: Union[str, pathlib.Path],
        callback: Optional[Callable] = None,
    ):
        """
        Supervises an entire digital worker run and generates a work report
        and drift report for the run from the digital worker's manifest.

        You can optionally specify a callback function that will be run when
        this context is finished via the `callback` param in the constructor. A
        callback is a function that is invoked with three parameters: the
        current context (as the `MainContext` instance), the `Report`, and
        the `DriftReport` generated from this digital worker's run.

        For example:

        ```python
        def print_work_report(
            ctx: MainContext,
            work_report: Report,
            drift_report: DriftReport
        ):
            print(work_report.__json__())

        def main()
            # ...
            pass

        with supervise(callback=print_work_report):
            main()
        ```

        Args:
            report_builder (ReportBuilder): Where step reports will be sent to
            recorder (Recorder): Where messages and data logs will be sent to
            manifest (str): The digital worker's manifest definition
            output_dir (str): Where the work report and drift report will
                be written to
            callback (callable, optional): a function that is invoked with three
                parameters: the current context (as the `MainContext` instance),
                the `Report`, and the `DriftReport` generated from this digital
                worker's run.
        """
        self.report_builder = report_builder
        self.recorder = recorder
        self.manifest_path = pathlib.Path(manifest)
        self.output_path = pathlib.Path(output_dir)
        self.callback = callback

    def __enter__(self) -> MainContext:
        """
        Logic for when this context is first started.

        Returns:
            MainContext: This instance.
        """
        if self.manifest_path.exists():
            try:
                _ = Manifest.from_file(self.manifest_path)
            except Exception as e:
                logger.exception("warning: could not read manifest")
        else:
            logger.warning(f"{self.manifest_path=} does not exist")
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """
        Runs when the context is about to close, whether caused
        by a raised Exception or now.

        Returns:
            bool: True if the parent caller should ignore the
                Exception raised before entering this function
                (if any), False otherwise.
        """
        if exc_type:
            self.report_builder.status = WorkerStatus.ERROR
        else:
            self.report_builder.status = WorkerStatus.COMPLETE

        work_report = self.report_builder.to_report()

        # Create the output directory if it doesn't already exist
        self.output_path.mkdir(parents=True, exist_ok=True)

        # Write the work report
        work_report_path = self._safe_report_path(file_prefix="run-report")
        work_report.write(work_report_path)

        # Write the manifest back out as a JSON file
        manifest_json_path = self.output_path / "manifest.json"
        Manifest.yaml_to_json(self.manifest_path, manifest_json_path)

        # Run the user-defined callback
        if self.callback:
            self.callback(self, work_report)

        # Explicitly return false so that the parent caller above
        # this context sees the exception (if any)
        return False

    def _safe_report_path(self, file_prefix: str) -> pathlib.Path:
        """
        A ``pathlib.Path`` instance that points to a new work report writable
        location that is safe across all OSes.

        Returns:
            pathlib.Path: The path to the new report to be written.
        """
        timestamp = datetime.datetime.now().strftime("%Y-%m-%dT%H%M%S")
        filename = f"{file_prefix}-{timestamp}.json"

        # Remove any characters from the timestamp that OSes don't like
        invalid_chars = [":", "*", "?", '"', "<", ">" "|", "'"]
        for char in invalid_chars:
            filename = filename.replace(char, "_")

        return self.output_path / filename



from __future__ import annotations

import datetime
import json
import math
import pathlib
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import isodate

from supervisor.__version__ import __version__
from supervisor.manifest import StepStatus
from supervisor.recorder import DataLog, MessageLog
from supervisor.reporting.record import Record


@dataclass
class StepReport:
    step_id: str
    status: StepStatus
    start_time: datetime.datetime
    end_time: datetime.datetime
    duration: datetime.timedelta
    args: Dict[str, Any] = field(default_factory=dict)
    outputs: Optional[Dict] = None
    message_log: MessageLog = field(default_factory=list)
    data_log: DataLog = field(default_factory=list)
    record: Optional[Record] = None

    def __json__(self) -> Dict[str, Any]:
        return {
            "step_id": self.step_id,
            "step_status": self.status.value,
            "start_time": self.start_time.isoformat(),
            "end_time": self.end_time.isoformat(),
            "duration": isodate.duration_isoformat(self.duration),
            "duration_in_ms": _time_delta_to_ms(self.duration),
            "args": self.args,
            "returned": self.outputs,
            "data_log": self.data_log,
            "message_log": self.message_log,
            "record": self.record.__json__() if self.record else None,
        }


class WorkerStatus(str, Enum):
    """The final status of the digital worker run."""

    COMPLETE = "complete"
    WARNING = "warning"
    ERROR = "error"


@dataclass
class Report:
    supervisor_version = str(__version__)
    workflow: List[StepReport] = field(default_factory=list)
    start_time: datetime.datetime = field(default=datetime.datetime.utcnow())
    end_time: datetime.datetime = field(default=datetime.datetime.utcnow())
    duration: datetime.timedelta = field(default=datetime.timedelta(seconds=0))
    status: Optional[WorkerStatus] = None

    def __json__(self) -> Dict[str, Any]:
        return {
            "supervisor_version": self.supervisor_version,
            "workflow": [step.__json__() for step in self.workflow],
            "start_time": self.start_time.isoformat(),
            "end_time": self.end_time.isoformat(),
            "duration": isodate.duration_isoformat(self.duration),
            "duration_in_ms": _time_delta_to_ms(self.duration),
            "status": self.status.value,
        }

    def write(self, filename: Union[str, pathlib.Path]) -> None:
        """
        Write the report as a JSON object to a file.

        Args:
            filename: Where to write the file.
        """
        path = (
            filename if isinstance(filename, pathlib.Path) else pathlib.Path(filename)
        )

        with path.open("w") as out:
            report_dict = self.__json__()
            json.dump(report_dict, out)


def _time_delta_to_ms(delta: datetime.timedelta) -> int:
    """
    Converts a timedelta into integer milliseconds.

    Args:
        delta: The time duration.

    Returns:
        The millisecond conversion of the timedelta as an int rounded up.
    """
    return math.ceil(delta.total_seconds() * 1000)



from __future__ import annotations

import json
import pathlib
from collections import defaultdict
from enum import Enum
from typing import Any, Dict, List, Optional, TypeVar, Union

import pydantic
import yaml
from pydantic import validator
from pydantic_yaml import YamlModel


class StepType(str, Enum):
    """
    StepType — by default, steps run in sequential order.

    A conditional step will determine the _next step at runtime based on the
    step_outputs of that step.

    If a step defines a `when_true` or `when_false` (a string of the next
    step id) then that step is considered conditional. After a conditional step
    is executed, the workflow iterator will compare the return value to
    the proceed_to[bool] value. If no value is provided to those properties,
    the next step in the sequence is executed.
    """

    SEQUENTIAL = "sequential"
    CONDITIONAL = "conditional"
    BRANCH = "branch"
    # Use "unknown" for steps indicated by the dev that aren't in the manifest
    UNKNOWN = "unknown"


class StepStatus(str, Enum):
    """
    StepStatus — The status of the current state of the step.

    READY: The step is locked and loaded
    RUNNING: The step is currently running
    COMPLETE: The step has completed
    RETRY: The step has failed, and will be retried
    WARNING: The step finished, but requires attention
    ERROR: The step has failed
    """

    READY = "ready"
    RUNNING = "running"
    COMPLETE = "complete"
    RETRYING = "retrying"
    WARNING = "warning"
    ERROR = "error"


StepId = TypeVar("StepId", bound=str)


class Step(YamlModel):
    """
    A step to execute in a manifest workflow.

        step_id (str): The ID of the step.
        title (str): The name of the step.
        description (str): A description of the step.
        steps (List[Step]): A list of children steps. Empty list if no children.
        when_false (str, optional): The next step ID if the previous step
            returned `False`. Defaults to None.
        when_true (str, optional): The next step ID if the previous step
            returned `True`. Defaults to None.
    """

    step_id: str
    step_type: StepType
    title: str
    description: Optional[str]
    steps: Optional[List[Step]] = None
    branches: Optional[Dict[Any, Any]]
    when_false: Optional[str]
    when_true: Optional[str]

    # PyCharm can't recognize that `@validator` creates a class method
    # noinspection PyMethodParameters
    @validator("steps", pre=True, always=True)
    def set_steps(cls, steps):
        return steps or []

    @classmethod
    def to_flat_dict(cls, steps: List[Step]) -> Dict[str, Step]:
        """
        Creates a flat dictionary of a list of steps, where each step's children
        is added to the flat dictionary.

        Args:
            steps (List[Step]): The steps to flatten.

        Returns:
            Dict[str, Step]: The steps as a flat dict. The keys are the step IDs
                and the values are the ``Step``s.
        """

        flat = {}
        for step in steps:
            flat[step.step_id] = step
            child_values = Step.to_flat_dict(step.steps)
            flat.update(child_values)

        return flat

    @property
    def has_children(self) -> bool:
        """
        Says whether this instance has children ``Step``s.

        Returns:
            bool: True if this instance has children, False otherwise.

        """
        return True if self.steps and len(self.steps) else False


class Manifest(YamlModel):
    """
    A digital worker manifest, typically read from a `manifest.yaml` file.
    Attributes:
        uid (str): The ID of the manifest.
        name (str): The name of the manifest.
        description (str): Manifest description.
        author (str): Who wrote the manifest.
        source (str): Where the manifest came from.
        workflow (List[Step]): The list of steps to execute.
        steps_by_id (dict): The same ``Step``s as `self.workflow`, but as a
            dictionary, where the keys are the ``Step`` IDs.
    """

    uid: str
    name: str
    description: str
    author: Optional[str]
    source: str
    workflow: List
    steps_by_id: Optional[Dict]  # This will get set after model initialization

    def set_steps_by_id(self) -> None:
        """
        After loading from a YAML/dict, this function creates a dictionary
        that maps step IDs to steps.
        """
        self.steps_by_id = Step.to_flat_dict(self.workflow)

    @classmethod
    def from_file(cls, filename: Union[str, pathlib.Path]) -> Manifest:
        """
        Creates an instance of this class from a YAML file.

        Args:
            filename: The name of the manifest .yaml file.

        Returns:
            Manifest: The new Manifest instance.
        """
        # Load the manifest using the base class importer
        try:
            manifest_in_flight = cls.parse_file(str(filename))
        except pydantic.error_wrappers.ValidationError as ex:
            errors = ex.errors()
            error_message = f"Invalid manifest file with {len(errors)} validation error(s) at the root level\n{_pretty_errors_string(errors)}"
            raise ValueError(error_message) from ex

        # The base importer loads the steps as dicts, so convert them into
        # proper steps
        step_dicts = manifest_in_flight.workflow
        final_steps: List[Step] = cls._load_manifest_file_steps(step_dicts)
        manifest_in_flight.workflow = final_steps
        manifest_in_flight.set_steps_by_id()

        return manifest_in_flight

    @classmethod
    def yaml_to_json(
        cls,
        manifest_path: Union[str, pathlib.Path],
        json_path: Union[str, pathlib.Path],
    ) -> None:
        """
        Converts a YAML manifest into a JSON manifest by converting the raw YAML
        contents into a JSON dictionary. This function is "dumb" and just writes
        the direct-to-JSON version of the YAML file.

        Args:
            manifest_path: The manifest YAML to consume.
            json_path: The output file to write the JSON payload to.
        """
        with manifest_path.open("r") as f:
            manifest_yaml = yaml.safe_load(f)

        with pathlib.Path(json_path).open("w") as json_path:
            json.dump(manifest_yaml, json_path)

    def __json__(self) -> dict[str, Any]:
        """
        Returns the values of this instance as a dictionary that are needed
        for a report.
        Returns:
            dict: The JSON-able dictionary (of only primitives) representation
                of this instance.
        """
        # Only return the keys we want in the report
        keys = ["uid", "name", "description", "author", "source"]
        value = {k: v for k, v in self.__dict__.items() if k in keys}
        return value

    # Private implementation

    @classmethod
    def _load_manifest_file_steps(cls, step_dicts: List[dict]) -> List[Step]:
        """
        Creates a list of Step instances based on a list of steps in dictionary
        format (usually loaded from a raw YAML file).

        Args:
            step_dicts: The steps as dictionaries to parse.

        Returns:
            List[Step]: The new Step instances.

        """
        successful_steps = []
        # Keep track of errors for specific step IDs, and also errors for any
        # steps that don't have a step ID
        errors_for_step_ids: dict[str, List[Any]] = defaultdict(list)
        errors_for_unknown_steps: List[Any] = []

        # Try parsing each step
        for step_dict in step_dicts:
            try:
                new_step = Step.parse_obj(step_dict)
                successful_steps.append(new_step)
            except pydantic.ValidationError as ex:
                errors = ex.errors()
                if "step_id" in step_dict:
                    _id = step_dict["step_id"]
                    errors_for_step_ids[_id].extend(errors)
                else:
                    errors_for_unknown_steps.append(errors)

        # Back out if there's no errors
        if not (errors_for_step_ids or errors_for_unknown_steps):
            return successful_steps

        # Otherwise, build the pretty print error string
        message = "Manifest step(s) are invalid"
        for step_id, errors in errors_for_step_ids.items():
            message += f"\nstep id: {step_id}\n"
            message += _pretty_errors_string(errors, indent=4)

        for index, step_errors in enumerate(errors_for_unknown_steps):
            message += f"\nunknown step {index + 1} (no step ID found)\n"
            message += _pretty_errors_string(step_errors, indent=4)

        raise ValueError(message)


def _pretty_errors_string(errors: list, indent: int = 0) -> str:
    """
    A custom string formatter for pydantic errors.

    See Also:
        The original implementation of this (with a slightly different format)
        is here - https://github.com/samuelcolvin/pydantic/blob/8846ec4685e749b93907081450f592060eeb99b1/pydantic/error_wrappers.py#L82-L83
    """

    def pretty_error(error: dict) -> str:
        loc_str = " -> ".join(str(ll) for ll in error["loc"])
        spacer = " " * indent
        return f'{spacer}{loc_str}\n    {spacer}{error["msg"]}'

    return "\n".join(pretty_error(err) for err in errors)



from __future__ import annotations

import datetime
import time
from dataclasses import dataclass
from typing import Optional


@dataclass
class TimerResult:
    """
    The start time, end time, and duration of running a timer.
    """

    start: datetime.datetime
    end: datetime.datetime
    duration: datetime.timedelta


@dataclass
class Timer:
    """
    Times an action by calling `start()` and `end()`.
    """

    start_time: Optional[datetime.datetime] = None
    end_time: Optional[datetime.datetime] = None
    duration: Optional[datetime.timedelta] = None
    _perf_counter_start_value: Optional[float] = None

    def start(self) -> None:
        """
        Starts the timer.
        """
        self._perf_counter_start_value = time.perf_counter()
        self.start_time = datetime.datetime.now()

    def end(self) -> TimerResult:
        """
        Ends the timer and records the duration.

        Returns:
            TimerResult: How long the timer ran and when it started.
        """
        # Find end time
        end_perf_counter = time.perf_counter()
        self.end_time = datetime.datetime.now()

        # Find duration
        duration_seconds = end_perf_counter - self._perf_counter_start_value
        self.duration = datetime.timedelta(seconds=duration_seconds)

        return TimerResult(
            start=self.start_time, end=self.end_time, duration=self.duration
        )



from __future__ import annotations

import datetime
import time
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from supervisor.manifest import StepId, StepStatus
from supervisor.recorder import DataLog, MessageLog
from supervisor.reporting.record import Record
from supervisor.reporting.report import Report, StepReport, WorkerStatus
from supervisor.reporting.status import Status
from supervisor.reporting.timer import Timer


@dataclass
class StepReportBuilder:
    """
    Builds a dynamic digital worker step.
    """

    step_id: str
    start_time: datetime.datetime
    end_time: datetime.datetime
    duration: datetime.timedelta
    status: StepStatus
    args: Dict[str, Any] = field(default_factory=dict)
    outputs: Optional[Dict] = None
    message_log: MessageLog = field(default_factory=list)
    data_log: DataLog = field(default_factory=list)
    record: Optional[Record] = None

    def to_report(self) -> StepReport:
        """
        A final report on this step's execution.

        Returns:
            StepReport: A final report on this step's execution.
        """

        # Build the report
        return StepReport(
            step_id=self.step_id,
            status=self.status,
            start_time=self.start_time,
            end_time=self.end_time,
            duration=self.duration,
            args=self.args,
            outputs=self.outputs,
            message_log=self.message_log,
            data_log=self.data_log,
            record=self.record,
        )


RecordId = str


@dataclass
class ReportBuilder:
    """
    A work report builder that creates a new work report as a digital worker
    is executed.
    """

    timer: Timer = field(default_factory=Timer)
    workflow: List[StepReportBuilder] = field(default_factory=list)
    timer_start: float = time.perf_counter()
    status: Optional[WorkerStatus] = None
    # These steps will be overridden with the specified status when the
    # `Report` is written
    _step_statuses_to_override: Dict[StepId, StepStatus] = field(default_factory=dict)
    _record_statuses_to_override: Dict[StepId, Dict[RecordId, Status]] = field(
        default_factory=lambda: defaultdict(dict)
    )

    def __post_init__(self):
        self.timer.start()

    def fail_step(self, step_id: str) -> None:
        """Override a step to be in the `StepStatus.ERROR` state."""
        self.set_step_status(step_id=step_id, status=StepStatus.ERROR)

    def set_step_status(self, step_id: str, status: Union[StepStatus, str]) -> None:
        """Override a step to be in the status of `status`"""
        # Convert the status to the correct type if necessary
        safe_status = StepStatus(status)
        self._step_statuses_to_override[step_id] = safe_status

    def set_record_status(
        self, step_id: str, record_id: str, status: Union[Status, str]
    ) -> None:
        """
        Override a record to be in the status of `status`.

        Args:
            step_id: The step id a specific step that contains this record.
            record_id: The id of the record to override.
            status: The status to override the record to.
        """
        # Convert the status to the correct type if necessary
        safe_status = Status(status)
        self._record_statuses_to_override[step_id][record_id] = safe_status

    def to_report(self) -> Report:
        """
        Convert supervisor workflow to work report.

        Returns:
            Report: The finalized work report.
        """
        timed = self.timer.end()

        # Update step reports with any overridden statuses
        steps = [x.to_report() for x in self.workflow]
        for step_report in steps:
            # Update step statuses
            if step_report.step_id in self._step_statuses_to_override:
                new_status = self._step_statuses_to_override[step_report.step_id]
                step_report.status = new_status
            # Update the record statuses
            if step_report.step_id in self._record_statuses_to_override:
                records = self._record_statuses_to_override[step_report.step_id]
                if step_report.record.record_id in records:
                    new_status = records[step_report.record.record_id]
                    step_report.record.status = new_status

        return Report(
            start_time=timed.start,
            end_time=timed.end,
            duration=timed.duration,
            workflow=steps,
            status=self.status,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['supervisor', 'supervisor.reporting']

package_data = \
{'': ['*']}

install_requires = \
['chevron>=0.14.0,<0.15.0',
 'isodate>=0.6.1,<0.7.0',
 'pre-commit>=2.17.0,<3.0.0',
 'pyconfs>=0.5.5,<0.6.0',
 'pydantic-yaml>=0.6.3,<0.7.0',
 'pydantic>=1.8.2,<2.0.0',
 'pyyaml>=5.4.1']

setup_kwargs = {
    'name': 'thoughtful',
    'version': '1.3.0',
    'description': 'Supervisor is a Workflow Engine for Digital Workers that generates a detailed telemetric log at runtime called a Work Report',
    'long_description': '# 👷 Supervisor\n\n<img\n  title="Supervisor"\n  alt="Supervisor — Github Header"\n  width="262px"\n  height="179.5px"\n  align="right"\n  src="https://user-images.githubusercontent.com/1096881/147704110-3116d1e3-c278-45d6-b99a-209faf2b17e0.png"\n/>\n\n> **:warning: NOTE**: *Supervisor* is quite new, so the API is likely to\n> change between now and 1.0. We welcome and encourage you to help shape future\n> development by [reporting issues][git:issues] and\n> [making suggestions][url:notion-feedback] 💖\n\n---\n<big>Supervisor is a <u>Workflow Engine</u> for Digital Workers that constructs\nand broadcasts a detailed and structured telemetric log, called the <u>Work Report</u>.</big>\n\n[![pre-commit](https://github.com/thoughtful-automation/supervisor/workflows/pre-commit/badge.svg?event=push)](https://github.com/thoughtful-automation/supervisor/actions?query=workflow%3Apre-commit+event%3Apush)\n[![test](https://github.com/thoughtful-automation/supervisor/workflows/test/badge.svg?event=push)](https://github.com/thoughtful-automation/supervisor/actions?query=workflow%3Atest+event%3Apush)\n\n<small>Supervisor supports `Python ≥ 3.7.5`</small>\n\n:books: 👉️ **[Read the Documentation](DOCUMENTATION.md)**\n\n## Table of Contents\n\n- [Install](#install)\n- [Documentation](#documentation)\n- [Contributing](#contributing)\n- [Resources](#resources)\n\n## Install\n\nSupervisor is a private package hosted on CodeArtifact.\n\n1. Authenticate with `CodeArtifact`:\n\n   ```bash\n   aws codeartifact login \\\n     --tool pip \\\n     --repository thoughtful-automation \\\n     --domain thoughtful-automation \\\n     --domain-owner XXXXXXXXXXXX \\\n     --region us-east-1\n   ```\n\n2. Pip install\n\n   ```bash\n   pip install supervisor\n   ```\n\n   > or install a specific version: `pip install supervisor==0.4.0`\n\n## Documentation\n\nSee [DOCUMENTATION.md](DOCUMENTATION.md).\n\n## Contributing\n\nContributions to Supervisor are welcomed!\n\nTo get started, see the [contributing guide](CONTRIBUTING.md).\n\n## Resources\n\nLinks to related code, documentation, and applications.\n\n[**🖥 Empower**][url:dwm]\n\n> The digital Workforce Manager (*DWM*)\n\n[**👷 Supervisor**][url:supervisor] (this repo)\n\n> The Workflow Engine for Digital Workers that constructs\nand broadcasts a detailed and structured telemetric log, called the Work Report\n\n[**:robot: Foundry**][url:otto]\n\n> The initialization tool for Digital Workers.\n\n[**🔀 Prospector**][url:prospector]\n\n> The design tool for Digital Workers.\n\n[**:books: Schema Library**][url:schema-lib]\n\n  > The JSON-Schema-defined documents used to validate the **Manifest** and the\n  > runtime **Work Report**\n\n[**:eagle: Department of Digital\n  Labor**][url:dodl]\n\n> Documentation and Specifications for building Digital Workers in *TA\'s\n> ecosystem*, and **Empower**\n\n---\n\n<div align="center">\n\n  Made with ❤️ by\n\n  [![Thoughtful Automation](https://user-images.githubusercontent.com/1096881/141985289-317c2e72-3c2d-4e6b-800a-0def1a05f599.png)][url:ta]\n\n</div>\n\n<!--  Link References -->\n\n[url:ta]: https://www.thoughtfulautomation.com/\n[url:dwm]: https://app.thoughtfulautomation.com/\n[url:supervisor]: https://github.com/Thoughtful-Automation/supervisor\n[url:otto]: https://github.com/Thoughtful-Automation/otto\n[url:prospector]: https://github.com/Thoughtful-Automation/prospector\n[url:dodl]: https://github.com/Thoughtful-Automation/dodl\n[url:schema-lib]: https://github.com/Thoughtful-Automation/schemas\n[url:notion-feedback]:\n    https://www.notion.so/thoughtfulautomation/Feedback-Feature-Requests-5716a73769ea4e0cba398e921eab44b5\n[git:issues]: https://github.com/Thoughtful-Automation/supervisor/issues\n',
    'author': 'Thoughtful Automation',
    'author_email': 'engineering@thoughtfulautomation.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://thoughtfulautomation.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
